import warnings
from typing import Optional


def deprecated(new: str, old: Optional[str] = None):
    """
    Warn a user once that a function has been deprecated
    If old is not specified, func.__name__ is used as the old name
    """

    def outer(func):
        old_name = old.__name__ if old is None else old

        def inner(*args, **kwargs):
            warnings.warn(f"Use {new} instead of {old_name}", DeprecationWarning)
            return func(*args, **kwargs)

        return inner

    return outer



import logging

from claripy.vsa import StridedInterval

l = logging.getLogger("angr_tests")


def check_si_fields(si, stride, lb, ub):
    lb &= si.max_int(si.bits)
    ub &= si.max_int(si.bits)
    if si.stride != stride:
        return False
    if si.lower_bound != lb:
        return False
    if si.upper_bound != ub:
        return False
    return True


def test_division():
    # non-overlapping

    # simple case 1
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # <4>0x1[0x0, 0x1]
    assert check_si_fields(op1.sdiv(op2), 1, 0, 1)
    assert check_si_fields(op1.udiv(op2), 1, 0, 1)

    # simple case 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=1)
    # <4>0x1[0xa, 0xc]
    assert check_si_fields(op1.sdiv(op2), 1, 10, 12)
    assert check_si_fields(op1.udiv(op2), 1, 10, 12)

    # simple case 3
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-2)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=-2)
    # udiv:  <4>0x1[0x0, 0x1]
    # sdiv:  <4>0x1[0xc, 0x4]
    assert check_si_fields(op1.udiv(op2), 1, 0, 1)
    assert check_si_fields(op1.sdiv(op2), 1, 12, 4)

    # simple case 4 : Result should be zero
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
    # BOT
    assert op1.sdiv(op2).is_empty
    assert op1.udiv(op2).is_empty

    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=4, upper_bound=6)
    # udiv: <4>0x0[0x0, 0x0]
    # sdiv: <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.udiv(op2), 0, 0, 0)
    assert check_si_fields(op1.sdiv(op2), 0, 0, 0)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-3, upper_bound=-1)
    # sdiv: <4>0x1[0x1, 0x6]
    # udiv: <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.sdiv(op2), 1, 1, 6)
    assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    # one in 0-hemisphere and one in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
    # sdiv: <4>0x1[0xc, 0xf]
    # udiv: <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.sdiv(op2), 1, 12, 15)
    assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    # Overlapping

    # case a of figure 2
    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=6)
    # sdiv: <4>0x1[0x0, 0x3]
    # udiv: <4>0x1[0x0, 0x3]
    assert check_si_fields(op1.sdiv(op2), 1, 0, 3)
    assert check_si_fields(op1.udiv(op2), 1, 0, 3)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-7, upper_bound=-1)
    # sdiv: <4>0x1[0x0, 0x6]
    # udiv: <4>0x1[0x0, 0x1]
    assert check_si_fields(op1.sdiv(op2), 1, 0, 6)
    assert check_si_fields(op1.udiv(op2), 1, 0, 1)

    # case b Fig 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=7, upper_bound=5)
    # sdiv: <4>0x1[0x0, 0xf]
    # udiv: <4>0x1[0x0, 0xa]
    assert check_si_fields(op1.sdiv(op2), 1, 0, 15)
    assert check_si_fields(op1.udiv(op2), 1, 0, 10)

    # case c Fig 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=5)
    # sdiv: <4>0x1[0x0, 0xe]
    # udiv: <4>0x1[0x0, 0xa]
    assert check_si_fields(op1.sdiv(op2), 1, 0, 14)
    assert check_si_fields(op1.udiv(op2), 1, 0, 10)

    # Strided Tests

    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=4, upper_bound=6)
    # sdiv: <4>0x0[0x0, 0x0]
    # udiv: <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.sdiv(op2), 0, 0, 0)
    assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=-3, upper_bound=-1)
    # sdiv: <4>0x1[0x1, 0x6]
    # udiv: <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.sdiv(op2), 1, 1, 6)
    assert check_si_fields(op1.udiv(op2), 0, 0, 0)

    # Overlapping case 1
    op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
    op2 = StridedInterval(bits=4, stride=3, lower_bound=7, upper_bound=3)
    # sdiv: <4>0x1[0x9, 0x7]
    # udiv: <4>0x1[0x0, 0x9]
    assert check_si_fields(op1.sdiv(op2), 1, 9, 7)
    assert check_si_fields(op1.udiv(op2), 1, 0, 9)

    # Overlapping case 2
    op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=1, upper_bound=3)
    # sdiv: <4>0x1[0x1, 0xd]
    # udiv: <4>0x1[0x1, 0x9]
    assert check_si_fields(op1.sdiv(op2), 1, 1, 13)
    assert check_si_fields(op1.udiv(op2), 1, 1, 9)


def test_multiplication():
    # non-overlapping

    # simple case 1
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # <4>0x2[0x2, 0x6]
    assert check_si_fields(op1.mul(op2), 2, 2, 6)

    # simple case 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=1)
    # <4>0x1[0xa, 0xc]
    assert check_si_fields(op1.mul(op2), 1, 10, 12)

    # simple case 3
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-2)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=-2)
    # Stride should be 2.
    # NOTE: previous result was: <4>0x1[0x4, 0x0] which is wrong.
    # possible values of 1[3,e] * 0[e,e] on 4 bits are [a, 8, 6, 4, 2, 0, e, c]
    # in the previous SI 2 was not present.
    assert check_si_fields(op1.mul(op2), 2, 2, 0)

    # simple case 4 : Result should be zero
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
    # <4>0x0[0x0, 0x0]
    assert check_si_fields(op1.mul(op2), 0, 0, 0)

    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=4, upper_bound=6)
    # Result: <4>0x1[0x4, 0x2]
    assert check_si_fields(op1.mul(op2), 1, 4, 2)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-3, upper_bound=-1)
    # Result <4>0x1[0x4, 0x2]
    assert check_si_fields(op1.mul(op2), 1, 4, 2)

    # one in 0-hemisphere and one in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
    # TOP
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Overlapping

    # case a of figure 2
    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=6)
    # TOP
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-7, upper_bound=-1)
    # TOP
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # case b Fig 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=7, upper_bound=5)
    # <4>0x1[0x0, 0xf]
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # case c Fig 2
    op1 = StridedInterval(bits=4, stride=1, lower_bound=3, upper_bound=-6)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=5)
    # <4>0x1[0x0, 0xf]
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Strided Tests

    # Both in 0-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=3)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=4, upper_bound=6)
    # <4>0x1[0x4, 0x2]
    assert check_si_fields(op1.mul(op2), 1, 4, 2)

    # Both in 1-hemisphere
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-6, upper_bound=-4)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=-3, upper_bound=-1)
    # <4>0x1[0x4, 0x2]
    assert check_si_fields(op1.mul(op2), 1, 4, 2)

    # Overlapping case 1
    op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
    op2 = StridedInterval(bits=4, stride=3, lower_bound=7, upper_bound=3)
    # <4>0x1[0x0, 0xf]
    assert check_si_fields(op1.mul(op2), 1, 0, 15)

    # Overlapping case 2
    op1 = StridedInterval(bits=4, stride=2, lower_bound=3, upper_bound=-7)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=1, upper_bound=3)
    # TOP
    assert check_si_fields(op1.mul(op2), 1, 0, 15)


def test_subtraction():
    # Basic Interval Tests
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
    # Result should be TOP
    assert check_si_fields(op1.sub(op2), 1, 0, 15)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=6)
    # Result should be 1,[-5, 5]
    # print(str(op1.sub(op2)))
    assert check_si_fields(op1.sub(op2), 1, -5, 5)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # Result should be 1,[15, 5]
    assert check_si_fields(op1.sub(op2), 1, 15, 5)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # Result should be 1,[-4, 5]
    assert check_si_fields(op1.sub(op2), 1, -4, 5)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
    # Result should be 1,[1, 7]
    assert check_si_fields(op1.sub(op2), 1, 1, 7)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
    # Result should be 1,[2, 12]
    # print(str(op1.sub(op2)))
    assert check_si_fields(op1.sub(op2), 1, 2, 12)

    # Strided Tests
    op1 = StridedInterval(bits=4, stride=2, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
    # Result should be TOP
    assert check_si_fields(op1.sub(op2), 1, 0, 15)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=2, upper_bound=6)
    # Result should be 1,[11, 5]
    assert check_si_fields(op1.sub(op2), 1, 11, 5)


def test_add():
    # Basic Interval Tests
    op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
    # Result should be TOP
    assert check_si_fields(op1.add(op2), 1, 0, 15)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=6)
    # Result should be 1,[3, 13]
    assert check_si_fields(op1.add(op2), 1, 3, 13)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # Result should be 1,[3, 9]
    assert check_si_fields(op1.add(op2), 1, 3, 9)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=2, upper_bound=2)
    # Result should be 1,[0,9]
    assert check_si_fields(op1.add(op2), 1, 0, 9)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=0)
    # Result should be 1,[1,7]
    assert check_si_fields(op1.add(op2), 1, 1, 7)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=-5, upper_bound=-1)
    # Result should be 1,[-4, 6]
    assert check_si_fields(op1.add(op2), 1, -4, 6)

    # Strided Tests
    op1 = StridedInterval(bits=4, stride=2, lower_bound=-2, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=-6)
    # Result should be TOP
    assert check_si_fields(op1.add(op2), 1, 0, 15)

    op1 = StridedInterval(bits=4, stride=1, lower_bound=1, upper_bound=7)
    op2 = StridedInterval(bits=4, stride=2, lower_bound=2, upper_bound=6)
    # Result should be 1,[3, 13]
    assert check_si_fields(op1.add(op2), 1, 3, 13)


if __name__ == "__main__":
    # Addition tests
    l.info("Performing Add Tests")
    test_add()
    l.info("Performing Subtraction Tests")
    test_subtraction()
    l.info("Performing Multiplication Tests")
    test_multiplication()
    l.info("Performing Division Tests")
    test_division()
    print("[+] All Tests Passed")



import unittest
from abc import abstractmethod

import claripy
from collections import defaultdict

import time

NUMBER_SOLUTIONS_TO_COMPARE = 5


def solution_ast(var, val):
    if type(var) == claripy.ast.strings.String:
        return claripy.StringV(val)
    return (var == val).args[1]


def all_equal(vals):
    vals = tuple(vals)
    if not vals:
        return True
    v0 = vals[0]
    return all(v == v0 for v in vals)


class SmtLibSolverTestCongruency(unittest.TestCase):
    @unittest.skip("Skip these test for now because of a problem with pysmt")
    def get_solvers(self):
        solvers = [
            claripy.SolverStrings(backend=claripy.backend_manager.backends.smtlib_cvc4),
            claripy.SolverStrings(backend=claripy.backend_manager.backends.smtlib_z3),
            # claripy.SolverStrings(backend=claripy.backend_manager.backends.smtlib_z3str),
        ]
        return solvers

    def _get_models_it(self, solver, variables, constraints, n=None):
        if n is None:
            n = NUMBER_SOLUTIONS_TO_COMPARE

        vars = list(variables)
        results = solver.batch_eval(vars, n, extra_constraints=constraints)
        for r in results:
            yield map(lambda x: (x[0], solution_ast(*x)), zip(vars, r))

    def doublecheck_model_is_correct(self, csts, model):
        for c in csts:
            new = c
            for var, val in model:
                new = new.replace(var, val)
            self.assertTrue(new.is_true(), f"Either model is incomplete or wrong! Constraint: {c}, Model: {model}")

    def is_model_correct(self, csts, model):
        for c in csts:
            new = c
            for var, val in model:
                new = new.replace(var, val)
            if not c.is_true:
                return False
        return True

    def _collect_generic_solver_test_data(self, variables, constraints):
        solvers = self.get_solvers()
        vars = list(variables)

        results = defaultdict(dict)
        results["variables"] = variables
        results["constraints"] = constraints
        results["solvers"] = solvers
        for s in solvers:
            before_sat_check = time.time()
            results[s]["claimed_sat"] = s.satisfiable(extra_constraints=constraints)
            results[s]["solve_time"] = time.time() - before_sat_check
            results[s]["claimed_solutions"] = ()

            if results[s]["claimed_sat"]:
                before_eval = time.time()

                solutions = tuple(
                    self._get_models_it(s, variables, constraints=constraints, n=NUMBER_SOLUTIONS_TO_COMPARE)
                )

                results[s]["eval_time"] = time.time() - before_eval

                results[s]["claimed_solutions"] = solutions
                results[s]["num_requested_solutions"] = NUMBER_SOLUTIONS_TO_COMPARE
                results[s]["solution_correctness"] = tuple(self.is_model_correct(constraints, sol) for sol in solutions)

        return dict(results)

    def _generic_consistency_check(self, results):
        vs = results["variables"]
        cs = results["constraints"]

        for solver in results["solvers"]:
            solver_results = results[solver]
            if not solver_results["claimed_sat"]:
                continue
            for r in solver_results["claimed_solutions"]:
                self.doublecheck_model_is_correct(cs, r)

    def _generic_congruency_check(self, solvers, results):
        self.assertTrue(all_equal(results[s]["claimed_sat"] for s in solvers), "Solvers disagree on satisfiability")
        self.assertTrue(
            all_equal(len(results[s]["claimed_solutions"]) for s in solvers), "Solvers disagree on number of solutions!"
        )

    def test_congruency_1(self):
        recv_input = claripy.StringS("recv_input", 1024)
        constraints = []
        constraints.append(0 <= claripy.StrIndexOf(recv_input, claripy.StringV("\r\n\r\n"), 0, 32))

        results = self._collect_generic_solver_test_data((recv_input,), constraints)
        self._generic_consistency_check(results)
        self._generic_congruency_check(results["solvers"], results)

    def test_congruency_2(self):
        """
        (set-logic ALL)
        (declare-fun recv_input_4_1024 () String)
        (assert (< ( str.indexof recv_input_4_1024 "\r\n\r\n" 0) 0))
        (assert (<= 0 ( str.indexof recv_input_4_1024 " \r\n" 0)))
        (assert (<= 0 ( str.indexof recv_input_4_1024 " \r\n" (+ ( str.indexof recv_input_4_1024 " \r\n" 0) 1))))
        (assert (<= 0 ( str.indexof recv_input_4_1024 " \r\n" (+ ( str.indexof recv_input_4_1024 " \r\n" (+ ( str.indexof recv_input_4_1024 " \r\n" 0) 1)) 1))))
        (assert (= "GET"
        ( str.substr
            ( str.substr
                recv_input_4_1024
                (+
                    ( str.indexof recv_input_4_1024 " \r\n" (+ ( str.indexof recv_input_4_1024 " \r\n" 0) 1))
                    1)
                ( str.indexof
                    recv_input_4_1024
                    " \r\n"
                    (+ ( str.indexof recv_input_4_1024 " \r\n" (+ ( str.indexof recv_input_4_1024 " \r\n" 0) 1)) 1)
                )
            )
            10
            1014)))
        (check-sat)
        """

        recv_input = claripy.StringS("recv_input", 1024)
        constraints = []

        def field_sep_idx(s, start_idx=0):
            return claripy.StrIndexOf(s, claripy.StringV(" \r\n"), start_idx, 32)

        constraints.append(claripy.StrIndexOf(recv_input, claripy.StringV("\r\n\r\n"), 0, 32) < 0)
        sep_idx_1 = field_sep_idx(recv_input)
        sep_idx_2 = field_sep_idx(recv_input, start_idx=sep_idx_1 + 1)
        sep_idx_3 = field_sep_idx(recv_input, start_idx=sep_idx_2 + 1)
        sep_idx_4 = field_sep_idx(recv_input, start_idx=sep_idx_3 + 1)

        constraints.append(sep_idx_1 >= 0)
        constraints.append(sep_idx_2 >= 0)
        constraints.append(sep_idx_3 >= 0)

        sub_start = sep_idx_2 + 1
        sub_end = sep_idx_3

        sub_str = claripy.StrSubstr(recv_input, sub_start, sub_end)

        constraints.append(claripy.StrSubstr(sub_str, 10, 1014) == claripy.StringV("GET"))

        results = self._collect_generic_solver_test_data((recv_input,), constraints)
        self._generic_consistency_check(results)
        self._generic_congruency_check(results["solvers"], results)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(SmtLibSolverTestCongruency)
    unittest.TextTestRunner(verbosity=2).run(suite)



from common_backend_smt_solver import if_installed
from unittest import TestCase, main
import claripy

import logging

l = logging.getLogger("claripy.test.solver")


def raw_hybrid_solver(reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = claripy.SolverHybrid()

    x = claripy.BVS("x", 32, min=0, max=10, stride=2)
    y = claripy.BVS("y", 32, min=20, max=30, stride=5)

    # TODO: for now, the stride isn't respected in symbolic mode, but we'll fix that next.
    # until we do, let's add constraints
    s.add(x <= 10)
    s.add(x % 2 == 0)
    s.add(y >= 20)
    s.add(y <= 30)
    s.add((y - 20) % 5 == 0)
    s.add(x != 8)

    assert sorted(s.eval(x, 20, exact=False)) == [0, 2, 4, 6, 8, 10]
    assert sorted(s.eval(x, 20)) == [0, 2, 4, 6, 10]
    assert sorted(s.eval(y, 20, exact=False)) == [20, 25, 30]
    assert sorted(s.eval(y, 20)) == [20, 25, 30]

    # test approximate_first option
    s._approximate_first = True
    old_solve_count = claripy._backends_module.backend_z3.solve_count
    assert sorted(s.eval(x, 20)) == [0, 2, 4, 6, 8, 10]
    assert claripy._backends_module.backend_z3.solve_count == old_solve_count
    s._approximate_firsrt = False

    # now constrain things further so that the VSA overapproximates
    s.add(x <= 4)
    assert sorted(s.eval(x, 20, exact=False)) == [0, 2, 4]
    assert sorted(s.eval(x, 20)) == [0, 2, 4]

    s.add(y >= 27)
    assert s.eval(y, 20, exact=False) == (30,)
    assert s.eval(y, 20) == (30,)

    t = claripy.SolverHybrid()
    x = claripy.BVS("x", 32)
    t.add(x <= 10)
    print(t.eval(x, 80, exact=False))
    assert len(t.eval(x, 5, exact=False)) == 5
    assert len(t.eval(x, 5, exact=False)) == 5
    assert len(t.eval(x, 6, exact=False)) == 6
    assert len(t.eval(x, 8, exact=False)) == 8
    assert len(t.eval(x, 99, exact=False)) == 11


def raw_replacement_solver(reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    sr = claripy.SolverReplacement()
    x = claripy.BVS("x", 32)
    assert len(sr.eval(x, 10)) == 10
    sr.add_replacement(x, claripy.BVV(0x101, 32))
    assert sr.eval(x, 10) == (0x101,)

    y = claripy.BVS("y", 32)
    sr.add([y + 1 == 200])
    assert (y + 1).cache_key in sr._replacements
    assert sr._replacement(y + 1) is claripy.BVV(200, 32)

    srb = sr.branch()
    assert len(srb.constraints) == len(sr.constraints)  # pylint:disable=no-member
    assert (y + 1).cache_key in sr._replacements
    assert sr._replacement(y + 1) is claripy.BVV(200, 32)

    sr = claripy.SolverReplacement()
    b = claripy.BoolS("b")
    assert sr._replacement(b) is b
    sr.add(claripy.Not(b))
    assert sr._replacement(b) is claripy.false

    sr = claripy.SolverReplacement(claripy.SolverVSA(), complex_auto_replace=True)
    x = claripy.BVS("x", 64)
    sr.add([x + 8 <= 0xFFFFFFFFFFFFFFFF])
    sr.add([x + 8 >= 0])
    assert sr._replacement(x) is not x


def raw_solver(solver_type, reuse_z3_solver):
    # bc = claripy.backends.BackendConcrete(clrp)
    # bz = claripy.backends.BackendZ3(clrp)
    # claripy.expression_backends = [ bc, bz, ba ]

    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = solver_type()

    s.simplify()

    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)

    l.debug("adding constraints")

    s.add(x == 10)
    s.add(y == 15)

    # Batch evaluation
    results = s.batch_eval([x + 5, x + 6, 3], 2)
    assert len(results) == 1
    assert results[0][0] == 15
    assert results[0][1] == 16
    assert results[0][2] == 3

    l.debug("checking")
    assert s.satisfiable()
    assert not s.satisfiable(extra_constraints=[x == 5])
    assert s.eval(x + 5, 1)[0] == 15
    assert s.solution(x + 5, 15)
    assert s.solution(x, 10)
    assert s.solution(y, 15)
    assert not s.solution(y, 13)

    shards = s.split()
    assert len(shards) == 2
    assert len(shards[0].variables) == 1
    assert len(shards[1].variables) == 1
    if isinstance(s, claripy.frontend_mixins.ConstraintExpansionMixin) or (
        isinstance(s, claripy.frontends.HybridFrontend)
        and isinstance(s._exact_frontend, claripy.frontend_mixins.ConstraintExpansionMixin)
    ):  # the hybrid frontend actually uses the exact frontend for the split
        assert {len(shards[0].constraints), len(shards[1].constraints)} == {
            2,
            1,
        }  # adds the != from the solution() check
    if isinstance(s, claripy.frontends.ReplacementFrontend):
        assert {len(shards[0].constraints), len(shards[1].constraints)} == {1, 1}

    # test result caching
    s = solver_type()
    s.add(x == 10)
    s.add(y == 15)
    assert not s.satisfiable(extra_constraints=(x == 5,))
    assert s.satisfiable()

    s = solver_type()
    # claripy.expression_backends = [ bc, ba, bz ]
    s.add(claripy.UGT(x, 10))
    s.add(claripy.UGT(x, 20))
    s.simplify()
    assert len(s.constraints) == 1
    # assert str(s.constraints[0]._obj) == "Not(ULE(x <= 20))"

    s.add(claripy.UGT(y, x))
    s.add(claripy.ULT(z, 5))

    # test that duplicate constraints are ignored
    old_count = len(s.constraints)
    s.add(claripy.ULT(z, 5))
    assert len(s.constraints) == old_count

    # print("========================================================================================")
    # print("========================================================================================")
    # print("========================================================================================")
    # print("========================================================================================")
    # a = s.eval(z, 100)
    # print("ANY:", a)
    # print("========================================================================================")
    # mx = s.max(z)
    # print("MAX",mx)
    # print("========================================================================================")
    # mn = s.min(z)
    # print("MIN",mn)
    # print("========================================================================================")
    # print("========================================================================================")
    # print("========================================================================================")
    # print("========================================================================================")

    print("CONSTRAINT COUNTS:", [len(_.constraints) for _ in s.split()])

    assert s.max(z) == 4
    assert s.min(z) == 0
    assert s.min(y) == 22
    assert s.max(y) == 2 ** y.size() - 1

    print("CONSTRAINT COUNTS:", [len(_.constraints) for _ in s.split()])

    ss = s.split()
    assert len(ss) == 2
    # if isinstance(s, claripy.frontend_mixins.ConstraintExpansionMixin):
    #   assert { len(_.constraints) for _ in ss } == { 3, 2 }

    # Batch evaluation
    s.add(y < 24)
    s.add(z < x)  # Just to make sure x, y, and z belong to the same solver, since batch evaluation does not support the
    # situation where expressions belong to more than one solver
    results = s.batch_eval([x, y, z], 20)
    assert set(results) == {
        (21, 23, 1),
        (22, 23, 3),
        (22, 23, 2),
        (22, 23, 4),
        (21, 22, 4),
        (21, 23, 4),
        (22, 23, 0),
        (22, 23, 1),
        (21, 22, 1),
        (21, 22, 3),
        (21, 22, 2),
        (21, 22, 0),
        (21, 23, 0),
        (21, 23, 2),
        (21, 23, 3),
    }

    # test that False makes it unsat
    s = solver_type()
    s.add(claripy.BVV(1, 1) == claripy.BVV(1, 1))
    assert s.satisfiable()
    s.add(claripy.BVV(1, 1) == claripy.BVV(0, 1))
    assert not s.satisfiable()

    # test extra constraints
    s = solver_type()
    x = claripy.BVS("x", 32)
    assert s.eval(x, 2, extra_constraints=[x == 10]) == (10,)
    s.add(x == 10)
    assert not s.solution(x, 2)
    assert s.solution(x, 10)

    # test signed min/max
    s = solver_type()
    x = claripy.BVS("x", 32)
    assert s.min(x, signed=True) == -0x80000000
    assert s.max(x, signed=True) == 0x7FFFFFFF
    s.add(claripy.ULE(x, 0x40000000) | claripy.UGE(x, 0xC0000000))
    assert s.min(x, signed=True) == -0x40000000
    assert s.max(x, signed=True) == 0x40000000

    # test result caching

    if isinstance(s, claripy.frontend_mixins.ModelCacheMixin):
        count = claripy._backends_module.backend_z3.solve_count

        s = solver_type()
        x = claripy.BVS("x", 32)
        s.add(x == 10)
        assert s.satisfiable()
        assert claripy._backends_module.backend_z3.solve_count == count
        assert s.eval(x, 1)[0] == 10
        assert claripy._backends_module.backend_z3.solve_count == count
        s.add(x == 10)
        s.add(x > 9)
        assert s.eval(x, 1)[0] == 10
        assert claripy._backends_module.backend_z3.solve_count == count

        y = claripy.BVS("y", 32)
        s.add(y < 999)
        assert s.satisfiable()
        assert claripy._backends_module.backend_z3.solve_count == count
        assert s.eval(y, 1)[0] == 0
        assert claripy._backends_module.backend_z3.solve_count == count


def raw_solver_branching(solver_type, reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = solver_type()
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    s.add(claripy.UGT(x, y))
    s.add(claripy.ULT(x, 10))

    assert s.eval(x, 1)[0] > 0

    t = s.branch()
    if isinstance(s, claripy.frontends.FullFrontend):
        assert s._tls.solver is t._tls.solver
        assert s._finalized
        assert t._finalized
    t.add(x == 5)
    # if isinstance(s, claripy.FullFrontend):
    #   assert t._solver is None

    s.add(x == 3)
    assert s.satisfiable()
    t.add(x == 3)
    assert not t.satisfiable()

    s.add(y == 2)
    assert s.satisfiable()
    assert s.eval(x, 1)[0] == 3
    assert s.eval(y, 1)[0] == 2
    assert not t.satisfiable()


def raw_combine(solver_type, reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s10 = solver_type()
    s20 = solver_type()
    s30 = solver_type()
    x = claripy.BVS("x", 32)

    s10.add(x >= 10)
    s20.add(x <= 20)
    s30.add(x == 30)

    assert s10.satisfiable()
    assert s20.satisfiable()
    assert s30.satisfiable()
    assert s10.combine([s20]).satisfiable()
    assert s20.combine([s10]).satisfiable()
    assert s30.combine([s10]).satisfiable()
    assert not s30.combine([s20]).satisfiable()
    assert s30.combine([s10]).eval(x, 1) == (30,)
    assert len(s30.combine([s10]).constraints) == 2


def raw_composite_solver(reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    # pylint:disable=no-member
    s = claripy.SolverComposite()
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)
    c = claripy.And(x == 1, y == 2, z == 3)
    s.add(c)
    assert len(s._solver_list) == 3
    assert s.satisfiable()

    s.add(x < y)
    assert len(s._solver_list) == 2
    assert s.satisfiable()

    s.simplify()
    assert len(s._solver_list) == 3
    assert s.satisfiable()

    s1 = s.branch()
    assert len(s1._solver_list) == 3

    s1.add(x > y)
    assert len(s1._solver_list) == 2
    assert not s1.satisfiable()
    assert len(s._solver_list) == 3
    assert s.satisfiable()

    s.add(claripy.BVV(1, 32) == claripy.BVV(2, 32))
    assert len(s._solver_list) == 3
    assert not s.satisfiable()

    ss = s.branch()
    assert len(ss._solver_list) == 3
    assert not ss.satisfiable()

    s = claripy.SolverComposite()
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)
    c = claripy.And(x == 1, y == 2, z > 3)
    s.add(c)

    if isinstance(s._template_frontend, claripy.frontend_mixins.ModelCacheMixin):
        assert len(s._solver_list) == 3
        count = claripy._backends_module.backend_z3.solve_count
        assert s.satisfiable()
        assert claripy._backends_module.backend_z3.solve_count == count + 3
        assert list(s.eval(x + y, 1)) == [3]
        assert claripy._backends_module.backend_z3.solve_count == count + 3


def raw_minmax(reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = claripy.Solver()
    x = claripy.BVS("x", 32)

    assert s.max(x) == 2**32 - 1
    assert s.min(x) == 0
    assert s.satisfiable()


def raw_composite_discrepancy(reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    a = claripy.BVS("a", 8)
    b = claripy.BVS("b", 8)
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)

    xy = x + y
    dst = claripy.BVV(0xBAAAAF50, 32) + xy
    constraints = []
    constraints.append(x <= 0x1)
    constraints.append(x != 0x0)
    constraints.append(claripy.SignExt(24, claripy.If(x > 0x0, a, 0)) != 0xA)
    constraints.append(x < 0x80)
    constraints.append(y <= 0x1)
    constraints.append(x == 0x1)
    constraints.append((0xBAAAAF50 + x) == 0xBAAAAF51)
    constraints.append(y != 0x0)
    constraints.append(claripy.SignExt(24, claripy.If(y > 0x0, b, 0)) != 0xA)
    constraints.append((x + y) < 0x80)
    constraints.append(z <= 0x1)
    constraints.append((x + y) == 0x2)

    sn = claripy.Solver()
    sc = claripy.SolverComposite()

    sn.add(constraints)
    sc.add(constraints)
    print(sn.max(dst), sc.max(dst))
    print(sn.min(dst), sc.min(dst))
    assert sn.max(dst) == sc.max(dst)
    assert sn.min(dst) == sc.min(dst)


def raw_ancestor_merge(solver, reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = solver()
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)

    s.add(x == 10)
    s.add(y == x)

    p = s.branch()
    q = s.branch()
    p.add(z == 1)
    q.add(z == 2)

    r = p.merge([q], [claripy.true, claripy.true])[-1]
    t = p.merge([q], [p.constraints[-1], q.constraints[-1]], common_ancestor=s)[-1]

    if not isinstance(r, claripy.frontends.CompositeFrontend):
        assert len(r.constraints) == 1
    assert len(t.constraints) == 3
    assert t.constraints[-1].variables == z.variables
    assert t.constraints[-1].op == "Or"
    assert len(t.constraints[-1].args) == 2


def raw_unsat_core(solver, reuse_z3_solver):
    claripy._backend_z3.reuse_z3_solver = reuse_z3_solver

    s = solver(track=True)
    x = claripy.BVS("x", 32)
    y = claripy.BVS("y", 32)
    z = claripy.BVS("z", 32)
    a = claripy.BVS("a", 32)

    s.add(x == y)
    s.add(x == 1)
    s.add(z != a)
    s.add(z == 2)
    s.add(a == 2)

    assert not s.satisfiable()
    unsat_core = s.unsat_core()
    assert len(unsat_core) == 3
    assert unsat_core[0] is not None
    assert unsat_core[1] is not None
    assert unsat_core[2] is not None


#
# Test Classes
#


class StandardTests(TestCase):
    @if_installed
    def test_composite_solver_with_strings(self):
        s = claripy.SolverComposite(
            template_solver_string=claripy.SolverCompositeChild(backend=claripy.backend_manager.backends.smtlib_cvc4)
        )
        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        z = claripy.BVS("z", 32)
        str_1 = claripy.StringS("sym_str_1", 1024)
        c = claripy.And(x == 1, y == 2, z == 3, str_1 == claripy.StringV("cavallo"))
        s.add(c)
        assert len(s._solver_list) == 4
        assert s.satisfiable()
        assert list(s.eval(str_1, 1)) == ["cavallo"]

    def test_minmax_with_reuse(self):
        raw_minmax(True)

    def test_minmax_without_reuse(self):
        raw_minmax(False)

    def test_composite_discrepancy_with_reuse(self):
        raw_composite_discrepancy(True)

    def test_composite_discrepancy_without_reuse(self):
        raw_composite_discrepancy(False)

    def test_model(self):
        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        s = claripy.Solver()

        s.add(x < 10)
        assert sorted(s.eval(x, 20)) == list(range(10))
        s.add(y == 1337)
        assert sorted(s.eval(x, 20)) == list(range(10))

    def test_unsatness(self):
        x = claripy.BVS("x", 32)

        s = claripy.Solver()
        s.add(x == 10)
        assert s.satisfiable()
        s.add(claripy.false)
        assert not s.satisfiable()

    def test_simplification_annotations(self):
        s = claripy.Solver()
        x = claripy.BVS("x", 32)

        s.add(x > 10)
        s.add(x > 11)
        s.add((x > 12).annotate(claripy.SimplificationAvoidanceAnnotation()))

        assert len(s.constraints) == 3
        s.simplify()
        assert len(s.constraints) == 2

    def test_zero_division_in_cache_mixin(self):
        # Bug in the caching backend. See issue #49 on github.
        num = claripy.BVS("num", 256)
        denum = claripy.BVS("denum", 256)
        e = claripy.BVS("e", 256)
        s = claripy.Solver()
        s.add(e == 8)
        assert s.satisfiable()
        s.add(claripy.If(denum == 0, 0, num / denum) == e)
        assert s.satisfiable()
        # As a bonus:
        s.add(num == 16)
        assert s.satisfiable()
        s.add(denum == 3)
        assert not s.satisfiable()

    def test_composite_solver_branching_optimizations(self):
        s = claripy.SolverComposite()
        w = claripy.BVS("w", 32)
        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        z = claripy.BVS("z", 32)

        s.add(w == 10)
        assert len(s._unchecked_solvers) == 1
        assert s.satisfiable()
        s.add(x == 10)
        assert len(s._unchecked_solvers) == 1

        s2 = s.branch()
        assert len(s2._unchecked_solvers) == 1
        assert s.satisfiable()
        assert len(s._unchecked_solvers) == 0
        assert len(s2._unchecked_solvers) == 1
        s.add(y == 10)
        assert len(s._unchecked_solvers) == 1
        assert len(s2._unchecked_solvers) == 1
        s2.add(z == 10)
        assert len(s._unchecked_solvers) == 1
        assert len(s2._unchecked_solvers) == 2
        assert s2.satisfiable()
        assert len(s._unchecked_solvers) == 1
        assert len(s2._unchecked_solvers) == 0

        s2.add(z == 12)
        assert not s2.satisfiable()
        assert not s2.satisfiable()

    def test_exhaustion(self):
        s = claripy.Solver()
        x = claripy.BVS("x", 32)
        s.add(x >= 19)
        print(s.min(x, extra_constraints=[x >= 20]))
        assert s.min(x) == 19

        s = claripy.Solver()
        x = claripy.BVS("x", 32)
        s.add(x <= 19)
        print(s.max(x, extra_constraints=[x <= 18]))
        assert s.max(x) == 19

    def test_cached_max(self):
        s = claripy.Solver()
        x = claripy.BVS("x", 32)
        assert not s.constraints
        assert s.max(x) == 0xFFFFFFFF
        assert len(s.constraints) == 1  # ConstraintExpansionMixin will add a new constraint
        assert s.max(x) == 0xFFFFFFFF  # calling it the second time, the cache should not give a different result

        s = claripy.Solver()
        y = claripy.BVS("y", 32)
        s.add(y == 8)
        assert s.eval(y, n=1)[0] == 8
        assert len(s.constraints) == 1
        assert s.max(x) == 0xFFFFFFFF
        assert s.max(x) == 0xFFFFFFFF


#
# Multi-Solver test base classes
#


class Base:
    def test_solver_with_reuse(self):
        raw_solver(self.solver, True)

    def test_solver_without_reuse(self):
        raw_solver(self.solver, False)

    def test_solver_branching_with_reuse(self):
        raw_solver_branching(self.solver, True)

    def test_solver_branching_without_reuse(self):
        raw_solver_branching(self.solver, False)

    def test_combine_with_reuse(self):
        raw_combine(self.solver, True)

    def test_combine_without_reuse(self):
        raw_combine(self.solver, False)

    def test_ancestor_merge_with_reuse(self):
        raw_ancestor_merge(self.solver, True)

    def test_ancestor_merge_without_reuse(self):
        raw_ancestor_merge(self.solver, False)


class UnsatCore(Base):
    def test_unsat_core_with_reuse(self):
        raw_unsat_core(self.solver, True)

    def test_unsat_core_without_reuse(self):
        raw_unsat_core(self.solver, False)


#
# Multi-Solver test base classes
#


class TestSolver(TestCase, UnsatCore):
    solver = claripy.Solver


class TestSolverReplacement(TestCase, Base):
    solver = claripy.SolverReplacement

    def test_replacement_solver_with_reuse(self):
        raw_replacement_solver(True)

    def test_replacement_solver_without_reuse(self):
        raw_replacement_solver(False)


class TestHybrid(TestCase, UnsatCore):
    solver = claripy.SolverHybrid

    def test_hybrid_solver_with_reuse(self):
        raw_hybrid_solver(True)

    def test_hybrid_solver_without_reuse(self):
        raw_hybrid_solver(False)


class TestComposite(TestCase, UnsatCore):
    solver = claripy.SolverComposite

    def test_composite_solver_with_reuse(self):
        raw_composite_solver(True)

    def test_composite_solver_without_reuse(self):
        raw_composite_solver(False)

    def test_composite_solver_child_solvers_transitive_closure(self):
        # https://github.com/angr/angr/issues/3604
        # the child solver must be merged from solvers for the transitive closure of all variable names

        s = claripy.SolverComposite()
        aa = claripy.BVS("aa", 32)
        bb = claripy.BVS("bb", 32)
        cc = claripy.BVS("cc", 32)
        dd = claripy.BVS("dd", 32)
        ee = claripy.BVS("ee", 32)

        s.add(dd[0:0] != 0)
        s.add(bb != 0)
        s.add(aa[15:0] != 0)
        s.add(claripy.If(aa[15:0] >= 0x1, claripy.BVV(0x1, 32), claripy.BVV(0x0, 32)) == 1)
        s.add(0xFFFFFFF + cc + bb == 0)
        s.add(claripy.If((claripy.If(bb == 0x0, bb, dd) | ee) == 0, claripy.BVV(1, 1), claripy.BVV(0, 1)) != 1)
        # print(s._solvers)

        s.add(dd == 0xB)
        s.simplify()
        # print(s._solvers)

        s.add(bb == 0x8004)
        s.simplify()
        # print(s._solvers)

        s.add(ee == 0)
        s.simplify()
        # print(s._solvers)

        bb_values = s.eval(bb, n=8)
        assert len(bb_values) == 1
        assert bb_values[0] == 0x8004


class TestSolverCacheless(TestCase, UnsatCore):
    solver = claripy.SolverCacheless


if __name__ == "__main__":
    main()
